#include <cmath>
#include <iomanip>
#include <iostream>
#include <sstream>

inline void check_cuda_error(const char* cudaop, const char* source,
       unsigned int line, cudaError_t error) {
   if (error != cudaSuccess) {
      std::cerr << cudaop << " at " << source << ":" << line
	 << " failed: " << cudaGetErrorString(error) << std::endl;
      exit(1);
   }
}

#define CHECK_CUDA(opname, ...) \
   check_cuda_error(#opname, __FILE__, __LINE__, opname(__VA_ARGS__))

unsigned int max_threads_per_block() {
   int device; CHECK_CUDA(cudaGetDevice, &device);
   struct cudaDeviceProp device_prop;
   CHECK_CUDA(cudaGetDeviceProperties, &device_prop, device);
   return device_prop.maxThreadsPerBlock;
}

// to be integrated function
__device__ double f(double x) {
   return 4 / (1 + x*x);
}

// numerical integration according to the Simpson rule
// for f over the i-th subinterval of [a,b]
__global__ void simpson(double a, double b, double* sums) {
   const int N = blockDim.x * gridDim.x;
   const unsigned int i = threadIdx.x + blockIdx.x * blockDim.x;
   double xleft = a + (b - a) / N * i;
   double xright = xleft + (b - a) / N;
   double xmid = (xleft + xright) / 2;
   sums[i] = (xright - xleft) / 6 * (f(xleft) + 4 * f(xmid) + f(xright));
}

char* cmdname;
void usage() {
   std::cerr << "Usage: " << cmdname << " [# intervals]" << std::endl;
   exit(1);
}

int main(int argc, char** argv) {
   const double a = 0;
   const double b = 1;
   unsigned int N = 8192;

   cmdname = *argv++; --argc;
   if (argc > 0) {
      std::istringstream arg(*argv++); --argc;
      if (!(arg >> N) || N <= 0) usage();
   }
   if (argc > 0) usage();

   int blocksize = max_threads_per_block();
   if (blocksize > N) {
      blocksize = N;
   } else {
      if (N % blocksize != 0) {
	 std::cerr << cmdname << ": please select a multiple of "
	    << blocksize << std::endl;
	 exit(1);
      }
   }
   int nof_blocks = N / blocksize;

   double* cuda_sums;
   CHECK_CUDA(cudaMalloc, (void**)&cuda_sums, N * sizeof(double));
   simpson<<<nof_blocks, blocksize>>>(a, b, cuda_sums);

   double sums[N];
   CHECK_CUDA(cudaMemcpy, sums, cuda_sums, N * sizeof(double), cudaMemcpyDeviceToHost);
   CHECK_CUDA(cudaFree, cuda_sums);
   double sum = 0;
   for (int i = 0; i < N; ++i) {
      sum += sums[i];
   }
   std::cout << std::setprecision(14) << sum << std::endl;
   std::cout << std::setprecision(14) << M_PI << std::endl;
}
